"""
DISCO Jobs representation
"""
import re
import time
import uuid
import sys
import subprocess
from collections import defaultdict
from typing import Union

from .models import JobDetails, Task
from .asset import Asset
from .core import analytics_reporting, constants, exceptions,\
    is_venv, zip_tools
from .task import TaskResult
from .base_controller import BaseController


class Job(BaseController):
    """A job that runs on DISCO machines.

    Every `Job` object has its own Python script and data files, and runs
    independently on the cloud, until it produces a result.
    """

    _task_result_pattern = re.compile('^([^.]*?).zip$')
    TIMEOUT = 10 * constants.MINUTE

    # WARNING: Does not handle paging
    @classmethod
    def list_jobs(cls, limit=None, next_=None):
        """Show a list of all the jobs belonging to this user.

        Args:
            limit (int):
            next_:

        Returns:
            list(JobDetails)
        """

        res = cls.query('findJobsByProfile', limit=limit,
                        ownerId=cls.get_current_user(),
                        archived=False, next=next_)

        if res is None:
            return []

        results = res['findJobs']['results']

        return [JobDetails(result) for result in results]

    @classmethod
    def jobs_summary(cls):
        """Gets a summary of all job statuses.

        Returns:
            dict: Dictionary [str, int] of status->count
        """

        status_counts = defaultdict(int)

        # This will work when list jobs supports paging and `yields`
        for job_detail in cls.list_jobs():
            status_counts[job_detail.status] += 1
        return dict(status_counts)

    @classmethod
    def _get_venv_requirements(cls, cluster_id: str):
        if is_venv():
            freeze_command = [sys.executable, '-m', 'pip', 'freeze']

            process = subprocess.Popen(
                freeze_command,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, _ = process.communicate()
            requirements = stdout.decode("utf-8")
            # filtering out disco packages
            split_req = filter(lambda line: "disco" not in line and "iqoqo" not in line, requirements.splitlines())
            requirements = "\n".join(split_req)
            req_file_id = Asset().upload(
                'requirements.txt',
                requirements,
                cluster_id)
            return req_file_id

        print("Warning: Python virtual environment was not detected. "
              "Dis.co will not be able to automatically "
              "update your cloud environment when not using "
              "Python virtual environment.")
        return None

    @classmethod
    def create(  # pylint: disable-msg=too-many-arguments

            cls,
            script_file_id,
            input_file_ids: Union[str, list] = None,
            constants_file_ids: Union[str, list] = None,
            job_name=None,
            cluster_instance_type='s',
            cluster_id=None,
            auto_start=False,
            upload_requirements_file=True
    ):
        """Creates a new job.

        Args:
            script_file_id (str): The ID of the script file to run.
            input_file_ids: A list of IDs of files that will be used as
                standard input.
            constants_file_ids:  A list of IDs of constants files.
            job_name: Is a name you can give to your job. Leave empty
                to use a random string.
            cluster_instance_type: Is the size of instance used. Choose 'm'
                for a medium instance and 'l' for a large instance.
                The default is 's' for small.
            cluster_id: Specifies the ID of the cluster on which to run the
                job. Leave as `None` to run on DISCO's cluster.
            auto_start: Automatically start the job upon creation.
            upload_requirements_file (boolean): if True uploads a requirements file if in venv

        Returns:
            obj: The created job object.
        """

        constants_file_ids = constants_file_ids or []
        input_file_ids = input_file_ids or []

        venv_requirements = cls._get_venv_requirements(cluster_id)
        if upload_requirements_file and venv_requirements:
            constants_file_ids.append(venv_requirements)

        analytics_reporting.setup_analytics()

        job_meta = {
            'name': job_name or uuid.uuid4().hex,
            'scriptId': script_file_id,
            'inputIds': input_file_ids,
            'constantIds': constants_file_ids,
            'clusterId': cluster_id,
            'clusterInstanceType': cluster_instance_type
        }

        job_options = {
            'executeOnSubmit': auto_start
        }

        result = cls.query('createJob', args=job_meta, options=job_options)
        analytics_reporting.track('job_register', job_meta)
        return Job(result['createJob']['id'])

    def __init__(self, job_id):
        self.job_id = job_id
        super().__init__()

    def start(self):
        """Start the job.

        When you run `job.start()`, the DISCO server will queue the job for
        execution.

        Returns:
            obj: The job object.
        """
        return self.query('startJob', id=self.job_id)

    def stop(self):
        """Cancels a running job.

        When you run `job.stop()`, the DISCO server will stop running the job
        and return any results retrieved so far.
        """

        self.query('stopJob', id=self.job_id)

    def archive(self):
        """Archive the job, making it unusable."""
        self.query('archiveJob', id=self.job_id)

    def get_details(self):
        """
        Get details about the job.

        This includes its name, last activity, status and task states.

        Returns:
            JobDetails
        """
        response = self.query('fetchJob', id=self.job_id)

        return JobDetails(response['fetchJob'])

    def get_tasks(self, limit=None, next_=None):
        """
        Get job tasks

        Args:
            limit (int):
            next_:

        Returns:
            list(Task)
        """
        # WARNING: Does not handle paging
        response = self.query(
            'findTasksByJob',
            limit=limit,
            next=next_,
            jobId=self.job_id)

        results = response['findTasks']['results']
        return [Task(result) for result in results]

    def get_status(self):
        """Get the job's status.

        This is a `JobStatus` object, equal to one of
        `['Unknown',
          'Listed',
          'Working',
          'Stopped',
          'Done',
          'Failed',
          'Deleted']`.

          Returns:
              The status of the job.
        """
        return constants.JobStatus(self.get_details().status)

    def wait_for_finish(self, interval=5, timeout=TIMEOUT):
        """Wait for a job to finish.

        This means waiting until it's no longer in "listed" or "working"
        statuses.

        Args:
            interval (int): Interval in seconds to check if the job has
                finished running.
            timeout (int): Timeout in seconds.

        Returns:
            The status of the job.
        """
        return self.wait_for_status(
            *constants.JOB_TERMINAL_STATES,
            interval=interval, timeout=timeout
        )

    def wait_for_status(self, *expected_statuses, interval=5, timeout=TIMEOUT):
        """Wait for one of a job statuses.

        Args:
            *expected_statuses (str): List of expected job statuses.
            interval (int): Interval in seconds to check the job's status.
            timeout (int): Timeout in seconds.

        Returns:
            The status of the job.
        """
        start_time = time.time()
        timeout_time = start_time + timeout

        while time.time() <= timeout_time:
            status = self.get_status()
            if status in expected_statuses:
                return status

            time.sleep(interval)
        raise exceptions.TimeOutError

    def get_results(self, block=False, block_timeout=TIMEOUT):
        """Get the job's result.

        Args:
            block (bool): Pass `block=True` to first wait for the job
                to be completed.
            block_timeout (int): timeout in seconds.

        Returns:
        """
        if block:
            self.wait_for_finish(block_timeout)

        response = self.stream(f"/jobs/{self.job_id}/result")

        archive = zip_tools.unzip_in_memory(response)
        task_results = []
        for file_name, file_content in archive:
            if file_name == 'no-results-found.txt':
                continue

            match = self._task_result_pattern.match(file_name)
            if not match:
                continue

            task_id = match.group(1)
            task_results.append(
                TaskResult(
                    task_id,
                    zip_tools.unzip_in_memory(file_content)
                )
            )
        return task_results

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, repr(self.job_id))



#  Copyright (c) 2019 Samsung, Inc. and its affiliates.
#
#  This source code is licensed under the RESTRICTED license found in the
#  LICENSE file in the root directory of this source tree.

import glob
import inspect
import os
import sys
from zipfile import ZipFile
import decorator
import dill
import time
from disco import asset, Job
import requests
SCRIPT_PICKLE_RESULT = """
with open('run-result/result.pickle', 'wb') as f_result:
    dill.dump(result, f_result)   
"""

SCRIPT_HEADER_OPEN_ARGS = """
import sys, dill
from os import listdir
import os
os.environ['is_remote']='yes'
with open(sys.argv[1], "rb") as f_arg:
    arg = dill.load(f_arg)
"""

SCRIPT_HEADER_OPEN_PICKLED_MODULE = """
with open('customer-module.pickle', 'rb') as f_module:
    module = dill.load(f_module)
"""


def create_stand_alone_script(func):
    if not func:
        return None
    if not inspect.isfunction(func):
        return None
    source = inspect.getsourcefile(func)
    func_name = func.__name__
    with open(source, "r") as f:
        sourcecode = '#!/usr/bin/python3.6\n' if (
                sys.version_info >= (3, 0)) else '#!/usr/bin/python2.7\n'
        sourcecode += SCRIPT_HEADER_OPEN_ARGS
        sourcecode += f.read()
        sourcecode += str.format("result = {}(*arg)\n", func_name)
        sourcecode += SCRIPT_PICKLE_RESULT
    return sourcecode


def create_input_file(arg):
    return dill.dumps(arg, byref=False, recurse=True, protocol=0)


@decorator.decorator
def disco_job(func, name_of_job=None, *args, **kwargs):
    if os.environ.get("is_remote") == "yes":
        return func(*args, **kwargs)

    job_name = name_of_job if name_of_job else func.__name__ + str(time.time())
    script = create_stand_alone_script(func)
    input_file = create_input_file(args)

    # TODO: Enable a client to specify a cluster ID
    script_file_id = asset.upload_file(str(job_name) + '.py', script)
    input_file_id = asset.upload_file(str(job_name) + 'input.pickle',
                                      input_file)
    job = Job.create(script_file_id, [input_file_id], [], job_name)
    job.start()
    task_result = job.get_results(block=True)
    if task_result:
        for file_name, content in task_result[0].get_raw_result():
            if file_name == 'result.pickle':
                return dill.loads(content)
    else:
        print("no task result")
        return None


@disco_job
def add(x, y):
    return x + y





class BaseModel:
    """
    Base Model
    """

    def __init__(self, data):
        self._data = data or {}

    def __repr__(self):
        if hasattr(self, 'id'):
            return '%s(%s)' % (type(self).__name__, repr(self.id))  # pylint: disable=no-member

        return dict.__repr__(self._data)



"""
Base functionality for SDK and CLI
"""
from .gql import Authentication


class BaseController(Authentication):
    """Base controller that provides authenticated methods of calling `rest` ,
    `gql` and `stream` functions
    """

    def __init__(self):
        pass



import json
import os
import random
import string
import sys
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory

from disco.core import Credentials
import mock

random_str = string.ascii_lowercase + string.ascii_uppercase + string.digits


class RandomDataMixin:
    @classmethod
    def random_str(cls, prefix='', length=10):
        return prefix + ''.join(random.choice(random_str) for _ in range(length))

    @classmethod
    def random_int63(cls):
        return random.getrandbits(63)

    @classmethod
    def random_int32(cls):
        return random.getrandbits(32)

    @classmethod
    def random_int16(cls):
        return random.getrandbits(16)

    @classmethod
    def random_int(cls, minimum=0, maximum=sys.maxsize):
        return random.randint(minimum, maximum)

    random_from = random.choice
    random_bytes = os.urandom

    @classmethod
    def random_dict(cls, length=10, k_prefix='', v_prefix=''):
        return {cls.random_str(k_prefix): cls.random_str(v_prefix) for _ in range(length)}

    @classmethod
    def random_list(cls, prefix='', length=10):
        return [cls.random_str(prefix) for _ in range(length)]

    @classmethod
    def random_email(cls, prefix='', length=10):
        return f"{cls.random_str(prefix, length)}@{cls.random_str(prefix, length)}.{cls.random_str(prefix, 3)}"

    @classmethod
    def random_url(cls, prefix='', length=10, relative=False):
        if relative:
            return cls.random_str(prefix, length)

        return f"https://{cls.random_str(prefix, length)}.{cls.random_str(prefix, length)}.{cls.random_str(prefix, 3)}"


class AssertionsMixin:
    @classmethod
    def assert_not_called(cls, name):
        def _f(*args, **kwargs):
            raise AssertionError(f"fn: {name} was not supposed to be called, but was invoked with {args}, {kwargs}")

        return _f

    @classmethod
    def any_instance_of(cls, target_type):
        class Any(cls):
            def __eq__(self, other):
                return isinstance(other, target_type)

        return Any()

    @classmethod
    def any(cls):
        class Any(cls):
            def __eq__(self, other):
                return True

        return Any()


class BaseTest(RandomDataMixin, AssertionsMixin):

    @classmethod
    @contextmanager
    def _mock_auth(cls, apply=False):
        if apply:
            with mock.patch('disco.gql.authentication.Authentication.refresh') as _refresh:
                yield
        else:
            yield

    @classmethod
    def _reset_auth(cls):
        Credentials.username = None
        Credentials.token = None
        Credentials.credentials = (None, None)

    @classmethod
    @contextmanager
    def with_config_and_env(cls, config=None, env=None, authenticated=True):
        config = {} if config is None else config
        env = {} if env is None else env
        os_env = {k: v for k, v in os.environ.items() if not k.startswith('DISCO_')}
        os_env.update(env)
        cls._reset_auth()
        if authenticated:
            Credentials.username = cls.random_str('default_username')
            Credentials.token = cls.random_str('default_token')
            Credentials.credentials = (cls.random_str('default_cred_username'), cls.random_str('default_pass'))

        with TemporaryDirectory() as d:
            conf_path = Path(d) / f"{cls.random_str('config_')}.json"
            conf_path.write_text(json.dumps(config))
            with mock.patch('disco.core.constants.DISCO_CONFIG_PATH', conf_path):
                old_env = os.environ
                with cls._mock_auth(authenticated):
                    try:
                        os.environ = env

                        yield

                    finally:
                        os.environ = old_env
                        cls._reset_auth()



from disco.base_controller import BaseController
from disco.models import ClusterDetails


class Cluster(BaseController):
    """
    Cluster methods
    """

    @classmethod
    def list_clusters(cls, limit=None, next_=None):
        """Show a list of all the clusters applicable for this user.

                Args:
                    limit (int): pagination limit
                    next_: pagination next

                Returns:
                    list(ClusterDetails): List of the clusters belonging to this user.
                """

        res = cls.query('fetchProfileClusters', limit=limit,
                        id=cls.get_current_user(), next=next_)

        if res is None:
            return []
        return [ClusterDetails(cluster, res['fetchProfile']['defaultClusterId'] == cluster['id'])
                for cluster in res['fetchProfile']['clusters']]



#  Copyright (c) 2019 Samsung, Inc. and its affiliates.
#
#  This source code is licensed under the RESTRICTED license found in the
#  LICENSE file in the root directory of this source tree.

import os
import subprocess
import sys
import mock
import pytest

import disco
from disco.core import constants as const, exceptions
from disco.models import JobDetails
from .base_test import BaseTest

    
class TestJobs(BaseTest):

    def test_job_summary(self):
        with mock.patch('disco.Job.list_jobs') as list_jobs_mock:

            expected_statuses = ['Done', 'Working', 'Done', 'Failed', 'Failed']
            jobs_details = [JobDetails(dict(status=status)) for status in expected_statuses]
            list_jobs_mock.return_value = jobs_details

            jobs_summary = disco.Job.jobs_summary()

        assert jobs_summary == {'Done': 2, 'Failed': 2, 'Working': 1}

    def test_job_summary_empty(self):
        with mock.patch('disco.Job.list_jobs') as list_jobs_mock:
            list_jobs_mock.return_value = []

            jobs_summary = disco.Job.jobs_summary()

        assert jobs_summary == {}

    @BaseTest.with_config_and_env(authenticated=True)
    def test_get_tasks(self):

        job_id = self.random_str('job-id')

        task_id1 = self.random_str('task_id1')
        task_status1 = 'Success'
        task_duration1 = self.random_int()
        task_input_file_id1 = self.random_str('input_file_id1')
        artifact_ids1 = [self.random_str('artifactA_id1'), self.random_str('artifactB_id1')]

        task_id2 = self.random_str('task_id2')
        task_status2 = 'Running'
        task_duration2 = self.random_int()
        task_input_file_id2 = self.random_str('input_file_id2')
        artifact_ids2 = [self.random_str('artifactA_id2'), self.random_str('artifactB_id2')]

        with mock.patch('disco.Job.query') as graphql_mock:
            graphql_mock.return_value = {
                'findTasks': {
                    'results': [
                        {
                            'id': task_id1,
                            'status': task_status1,
                            'stats': {'duration': task_duration1},
                            'request': {'inputFile': {'id': task_input_file_id1}},
                            'result': {'artifactIds': artifact_ids1}
                        },
                        {
                            'id': task_id2,
                            'status': task_status2,
                            'stats': {'duration': task_duration2},
                            'request': {'inputFile': {'id': task_input_file_id2}},
                            'result': {'artifactIds': artifact_ids2}
                        },
                    ]
                }
            }

            tasks = disco.Job(job_id).get_tasks()

        assert isinstance(tasks, list)
        assert len(tasks) == 2

        [task1, task2] = tasks

        assert task1.id == task_id1
        assert task1.status == task_status1
        assert task1.duration == task_duration1
        assert task1.input_file_id == task_input_file_id1
        assert task1.artifact_ids == artifact_ids1

        assert task2.id == task_id2
        assert task2.status == task_status2
        assert task2.duration == task_duration2
        assert task2.input_file_id == task_input_file_id2
        assert task2.artifact_ids == artifact_ids2

    @BaseTest.with_config_and_env(authenticated=True)
    def test_list_jobs(self):

        job_id1 = self.random_str('job-id1')
        job_name1 = self.random_str('job-name1')
        job_status1 = 'Done'

        job_id2 = self.random_str('job-id2')
        job_name2 = self.random_str('job-name2')
        job_status2 = 'Failed'

        with mock.patch('disco.Job.query') as graphql_mock:
            graphql_mock.return_value = {
                'findJobs': {
                    'results': [
                        {
                            'id': job_id1,
                            'status': job_status1,
                            'request': {'meta': {'name': job_name1}},
                        },
                        {
                            'id': job_id2,
                            'status': job_status2,
                            'request': {'meta': {'name': job_name2}},
                        },
                    ]
                }
            }

            job_details = disco.Job.list_jobs()

        assert isinstance(job_details, list)
        assert len(job_details) == 2

        [job_detail1, job_detail2] = job_details

        assert job_detail1.id == job_id1
        assert job_detail1.name == job_name1
        assert job_detail1.status == job_status1
        assert job_detail1.tasks_summary is None

        assert job_detail2.id == job_id2
        assert job_detail2.name == job_name2
        assert job_detail2.status == job_status2
        assert job_detail2.tasks_summary is None

    @BaseTest.with_config_and_env(authenticated=True)
    def test_list_jobs_empty(self):
        with mock.patch('disco.Job.query') as graphql_mock:
            graphql_mock.return_value = None

            result = disco.Job.list_jobs()

        assert result == []

    @BaseTest.with_config_and_env()
    @mock.patch('disco.job.Job._get_venv_requirements', return_value=None)
    def test_return_id(self, _):
        expected_job_id = 'fake_job_id'
        with mock.patch('disco.gql.Query.send') as graphql_mock:
            graphql_mock.return_value = {
                'createJob': {
                    'id': expected_job_id
                }
            }

            job = disco.Job.create('fake_script_id')
            assert job.job_id == expected_job_id

    @mock.patch('disco.Job.get_status',
                mock.MagicMock(return_value=const.JobStatus.listed))
    def test_timeout(self):
        job = disco.job.Job('lucky')
        with pytest.raises(exceptions.TimeOutError):
            job.wait_for_finish(interval=1, timeout=0)

    @BaseTest.with_config_and_env()
    @mock.patch('disco.job.is_venv', return_value=True)
    def test__get_venv_requirements(self, _is_venv):
        with mock.patch('subprocess.Popen') as _popen, \
                mock.patch('disco.asset.Asset.upload') as _upload:
            res = '\n'.join([self.random_str() for _ in range(10)])
            bin_res = res.encode('utf-8')
            fid = self.random_str('requirements.txt_fid')
            cluster_id = self.random_str('cluster_id')

            _upload.return_value = fid
            _popen.return_value.communicate.return_value = (bin_res, self.random_bytes(10))

            res_fid = disco.Job._get_venv_requirements(cluster_id=cluster_id)

            assert res_fid == fid

            _popen.return_value.communicate.assert_called_once()
            _popen.assert_called_once_with(
                [sys.executable, '-m', 'pip', 'freeze'],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            _upload.assert_called_once()
            _is_venv.assert_called_once()



from disco.models import ClusterDetails
from tests.base_test import BaseTest


class TestClusterDetails(BaseTest):
    def test_cluster_create(self):
        cluster_id = self.random_str('cluster_id')
        cluster_name = self.random_str('cluster_name')
        account_id = self.random_str('id')

        cluster_data = {
            'id': cluster_id,
            'name': cluster_name,
            'type': 'aws',
            'cluster': {
                'region': 'us-west-2',
                'externalAccountId': account_id,
                'bucketName': 'disco-store-production'},
            'isLastConnectionCheckValid': True,
            'isActive': True,
            'enableCloudOffload': False
        }

        cluster = ClusterDetails(cluster_data, True)
        assert cluster.id == cluster_id
        assert cluster.name == cluster_name
        assert cluster.bucket_name == 'disco-store-production'
        assert cluster.type == 'aws'
        assert cluster.is_default
        assert cluster.is_active
        assert not cluster.cloud_offload_enabled
        assert cluster.is_last_connection_check_valid



import mock
import pkg_resources
from tests.base_test import BaseTest
from disco.core import utils


class TestUtils(BaseTest):
    @mock.patch('pkg_resources.get_distribution')
    def test_installed_version_ok(self, _get_distribution: mock.MagicMock):
        pkg = self.random_str('pkg')
        version = self.random_str('version')
        res = mock.MagicMock(version=version)

        _get_distribution.return_value = res
        assert utils.pkg_version(pkg) == version
        _get_distribution.assert_called_once_with(pkg)

    @mock.patch('pkg_resources.get_distribution')
    def test_installed_version_none(self, _get_distribution):
        pkg = self.random_str('pkg')

        _get_distribution.side_effect = pkg_resources.DistributionNotFound
        assert utils.pkg_version(pkg) is None
        _get_distribution.assert_called_once_with(pkg)



from mock import patch
from tests.base_test import BaseTest
from cli.utilities import  utils

class TestUtils(BaseTest):

    @patch('cli.utilities.utils._installed_version')
    def test_current_version_updated(self, mock_get_cli_installed_version):
        mock_get_cli_installed_version.return_value = "500.0.0"
        result = utils.is_update_needed()

        assert result == False

    @patch('cli.utilities.utils._installed_version')
    def test_current_version_not_updated(self, mock_get_cli_installed_version):
        mock_get_cli_installed_version.return_value = "1.0.0"
        result = utils.is_update_needed()

        assert result == True

    @patch('cli.utilities.utils._installed_version')
    def test_current_version_equal(self, mock_get_cli_installed_version):
        current_version = utils._latest_version()
        mock_get_cli_installed_version.return_value = current_version

        result = utils.is_update_needed()

        assert result == False


import errno
import glob
import os
import time

import click

from cli.command_utils import info_message, list_to_table, \
    error_message, success_message, verify_logged_in, exception_message
from disco import Job, upload_file
from pathlib import Path

from disco.core.exceptions import DiscoException
from disco.models import JobDetails

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS)
def job():
    """Manage jobs."""
    pass

@job.command(context_settings=CONTEXT_SETTINGS)
@click.option('-s', '--script', prompt='Enter path to script file', help='Local path to the script file to run.',
              type=click.Path(exists=True))
@click.option('-i', '--input', help='Local comma-separated path to the input file(s) for the script, (wildcards allowed).')
@click.option('-c', '--constants', help='Local path to the constants file that will be sent to each task.')
@click.option('-n', '--name', help='Job name.', prompt="Please enter job name")
@click.option('-r', '--run', help='Create and run the job.', is_flag=True)
@click.option('-w', '--wait', help='Wait for the job to finish.', is_flag=True)
@click.option('-cit', '--instance-type', help="Cluster instance type, options are: -s (small - default), -m (meduim), -l (large).",
              type=click.Choice(['s', 'm', 'l']), default='s') # todo: read form config: https://click.palletsprojects.com/en/7.x/options/
@click.option('-cid', '--cluster-id', help="Cluster id to run the job on. "
                                            "You can use the ‘cluster list’ command to view your clusters. "
                                            "In case no cluster-id specified, Dis.co runs the job "
                                            "on the default Dis.co cluster.")
@click.option('-e', '--repository', help='Repository name.')
@click.option('--debug-mode', hidden=True, is_flag=True)
@click.option('--dont-generate-req-file', hidden=True, is_flag=True)
def create(script, input, constants, name, run, wait, instance_type, cluster_id,
           repository, debug_mode, dont_generate_req_file):
    """
    Create a new job that can be launched immediately, or later.

    $ disco job create —n <job’s name> -s <script file path> -i <input file path> --run
    """
    if not verify_logged_in():
        return
    script_path = Path(script)
    if not (script_path.suffix in ['.sh', '.py']):
        error_message("Cannot use script file. Currently only Python and bash scripts are supported")
        return

    try:
        input_paths = get_file_list(input) if input else []
        const_paths = get_file_list(constants) if constants else []
    except FileNotFoundError as ex:
        error_message(str(ex))
        return

    try:
        const_file_ids, input_file_ids, script_file_id = upload_job_files(script, input_paths, const_paths, cluster_id)
        created_job = Job.create(script_file_id, input_file_ids, const_file_ids, name, instance_type,
                                 cluster_id, run, upload_requirements_file=not dont_generate_req_file)
        if run:
            info_message(' Job started')
        if wait:
            info_message(' Waiting for job to finish')
            created_job.wait_for_finish()
        display_job(created_job.job_id)
    except DiscoException as ex:
        if debug_mode:
            print(ex)
        exception_message(ex)


@job.command(context_settings=CONTEXT_SETTINGS)
@click.argument('job_id', metavar='<JOB_ID>')
def start(job_id):
    """
    Start an existing job.

    $ disco job start <job_id>

    You may use ‘job list’ command to get the job id.
    """
    if not verify_logged_in():
        return
    try:
        job = Job(job_id)
        job.start()
        success_message(f"Job {job_id} started")
    except DiscoException as ex:
        error_message("Error occurred while starting job")
        exception_message(ex)


@job.command('list', context_settings=CONTEXT_SETTINGS)
def list_jobs():
    """
    List all jobs and their status.

    $ disco job list

    Status Decription: \n

    Deleted\t The job failed\n\r
    Done\t The job is done\n
    Failed\t The job failed\n
    Listed\t The job is queued\n
    Stopped\t The job has stopped\n
    Unknown\t The job has no status\n
    Working\t The job is working\n
    """

    if not verify_logged_in():
        return
    try:
        job_list = Job.list_jobs()
        display_job_list(job_list)
    except DiscoException as ex:
        exception_message(ex)


@job.command(context_settings=CONTEXT_SETTINGS)
@click.argument('job_id', metavar='<JOB_ID>')
def stop(job_id):
    """
    Stop a running job.

    $ disco job stop <job_id>

    You may use ‘job list’ command to get the job id.
    """

    if not verify_logged_in():
        return
    try:
        job = Job(job_id)
        job.stop()
        success_message(f"Stopping job {job_id}")
    except DiscoException as ex:
        exception_message(ex)


@job.command(context_settings=CONTEXT_SETTINGS)
@click.argument('job_id', metavar='<JOB_ID>')
def archive(job_id):
    """
    Archive a specific job.

    $ disco job archive <job_id>

    You may use ‘job list’ command to get the job id.
    """

    if not verify_logged_in():
        return

    try:
        job = Job(job_id)
        job.archive()
        success_message(f"Job {job_id} was archived")
    except DiscoException as ex:
        error_message("Failed to archive job")
        exception_message(ex)


@job.command(context_settings=CONTEXT_SETTINGS)
@click.argument('job_id', metavar='<JOB_ID>')
def view(job_id):
    """
    View specific job's details.

    $ disco job view <job_id>

    You may use ‘job list’ command to get the job id.
    """

    if not verify_logged_in():
        return
    try:
        job = Job(job_id)
        details = job.get_details()
        display_job_details(details)
    except DiscoException as ex:
        error_message("Error occurred while getting job details")
        exception_message(ex)


@job.command(context_settings=CONTEXT_SETTINGS)
@click.argument('job_id', metavar='<JOB_ID>')
@click.option('-d', '--destination', prompt="Path for result files", help="The destination path to download the result files to.")
def download_results(job_id, destination):
    """
    Download results of a job.

    $ disco job download-results <job_id> -d <results files path>

    You may use ‘job list’ command to get the job id.
    """

    if not verify_logged_in():
        return
    dest_dir = Path(destination)
    if not dest_dir.exists():
        error_message(f"{destination} doesn't exist")
        return
    if not dest_dir.is_dir():
        error_message("Destination must be a folder")
        return

    try:
        task_results = Job(job_id).get_results(block=True)
        for task_result in task_results:
            task_dir = os.path.join(destination, f"{int(time.time() * 1000)}-{task_result.task_id}")
            safe_make_dirs(task_dir)
            task_result.write_files(task_dir)
            success_message("Results downloaded successfully")
    except DiscoException as ex:
        error_message(f"Error occurred while getting job {job_id} results")
        exception_message(ex)


def display_job_list(job_list):
    """Receives a list of jobs and prints them as a table"""
    tabulated_array = []
    for job in job_list:
        tabulated_array.append([job.id, job.name, job.status])
    displayable_table = list_to_table(tabulated_array, ["ID", "Name", "Status"])
    info_message(displayable_table)


def display_job(job):
    """Displays the job id on the screen"""
    info_message(f"Created job with id {job}")


def display_job_details(job: JobDetails):
    """Receives a job details and prints it to the screen"""
    info_message(f"Name: {job.name}")
    info_message(f"Status: {job.status}")
    info_message('Task summary')
    info_message('------------')
    info_message(f"Waiting: {job.tasks_summary.queued}")
    info_message(f"queued: {job.tasks_summary.queued}")
    info_message(f"running: {job.tasks_summary.running}")
    info_message(f"failed: {job.tasks_summary.failed}")
    info_message(f"success: {job.tasks_summary.success}")


def get_file_list(file_list_string):
    file_paths = []
    split_file_list = file_list_string.split(',')
    for file_path in split_file_list:
        if file_path.find('*') >= 0:
            files = glob.glob(file_path)
            files.sort(key=os.path.getmtime)
            file_paths += [Path(file) for file in files]
            continue
        path = Path(file_path)
        if not path.exists():
            raise FileNotFoundError(f"{file_path} doesn't exist")
        if path.is_dir():
            expanded_file = list(Path(file_path).glob('**/*'))
            if len(expanded_file) == 0:
                raise FileNotFoundError(f"Folder {file_path} is empty")
            file_paths += expanded_file
        else:
            file_paths.append(path)
    return file_paths


def upload_job_files(script, input_paths, const_paths, cluster_id):
    """

    Args:
        script: path to script file
        input_paths: list of paths to input file
        const_paths: list of paths to constant files
        cluster_id: cluster id

    Returns:
        DB Ids for constant files, input files and script file

    """
    script_file_id = upload_file(Path(script).name, Path(script), cluster_id)
    info_message(" Script file uploaded")
    input_file_ids = []
    for input_file in input_paths:
        input_file_ids.append(upload_file(input_file.name, input_file, cluster_id))
        info_message(f" Uploaded input file {input_file}")
    const_file_ids = []
    for const_file in const_paths:
        const_file_ids.append(upload_file(const_file.name, const_file, cluster_id))
        info_message(f" Uploaded constants file {const_file}")
    return const_file_ids, input_file_ids, script_file_id


def safe_make_dirs(dir):
    """

    Args:
        dir: directory to create

    Returns:

    """
    try:
        os.makedirs(dir)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise



"""
Credentials related functionality (get current, set and load from env/config
"""
import os
import json

from typing import Tuple

from . import constants, exceptions


class Credentials:
    """Simple skeleton for credentials."""
    token = None
    username = None
    credentials: Tuple[str, str] = (None, None)

    @classmethod
    def has_credentials(cls):
        """Checks if the credentials object contains credentials"""
        return None not in list(cls.credentials)

    @classmethod
    def from_config(cls) -> Tuple[str, str]:
        """Gets credentials from environment variables or configuration file

        Returns:
            tuple [str, str]:
        """
        env_mail = os.environ.get(constants.ENV_VAR_EMAIL_NAME, None)
        env_pass = os.environ.get(constants.ENV_VAR_PASSWORD_NAME, None)

        if None not in (env_mail, env_pass):
            return env_mail, env_pass

        if constants.DISCO_CONFIG_PATH.exists():
            disco_config_content = json.loads(
                constants.DISCO_CONFIG_PATH.read_text())

            return (disco_config_content.get('email'),
                    disco_config_content.get('password'))

        raise exceptions.NoCredentials

    @classmethod
    def update(cls, email, password, save_to_config=False):
        """Set the DISCO credentials to be used when communicating
        with the server.
        Pass `save_to_config=True` to save the credentials to
        `~/.disco_py`, so next time you wouldn't have to set them.
        No need to login after running `set_credentials`,
        login will be done automatically on-demand.

        Args:
            email (str):
            password (str):
            save_to_config (bool):
        """
        if cls.credentials == (email, password):
            return

        cls.credentials = (email, password)
        cls.token = None
        cls.username = None
        if save_to_config:
            json_content = json.dumps(
                {'email': email,
                 'password': password}
            )
            constants.DISCO_CONFIG_PATH.write_text(json_content)

    @classmethod
    def reset(cls):
        """Resets the credentials to none"""
        cls.update(None, None, save_to_config=True)
        cls.token = None


def get_credentials_from_config() -> Tuple[str, str]:
    """Legacy mapping to `Credentials.from_config` Gets credentials from
    environment variables or configuration file

    Returns:
        tuple [str, str]:
    """
    return Credentials.from_config()


def set_credentials(*args, **kwargs):
    """Legacy mapping to `Credentials.update`
    Set the DISCO credentials to be used when communicating with the server.
    Pass `save_to_config=True` to save the credentials to `~/.disco_py`, so
    next time you wouldn't have to set them.
    No need to login after running `set_credentials`, login will be done
    automatically on-demand.

    Args:
        *args:
        **kwargs:
    """
    return Credentials.update(*args, **kwargs)



from pathlib import Path
from time import sleep

from disco import Job
from .sdk_mocks import MockListJobsResponse, MockViewJobResponse, MockBadIdException
from mock import patch
from click.testing import CliRunner

from disco_cli import cli, setup_cli
from .cli_test_utils import output_message_includes

from tests.base_test import BaseTest


class TestJobCommands(BaseTest):

    def setup_class(self):
        setup_cli()

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    def test_job_list(self, is_logged_in_mock):
        """
        Tests job list
        Returns:

        """
        is_logged_in_mock.return_value = True
        with patch('disco.Job.list_jobs') as list_jobs_mock:
            list_jobs_mock.return_value = MockListJobsResponse
            runner = CliRunner()
            result = runner.invoke(cli, ['job', 'list'])
            assert result.exit_code == 0
            list_jobs_mock.assert_called()
            is_logged_in_mock.assert_called()
            assert output_message_includes(result, '| 5d66595208edfa000a250dda | Cool Humor       | Done     |')

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.list_jobs')
    def test_action_without_login(self, list_jobs_mock, is_logged_in_mock):
        list_jobs_mock.return_value = []
        is_logged_in_mock.return_value = False
        runner = CliRunner()
        result = runner.invoke(cli, ['job', 'list'])
        assert result.exit_code == 0
        assert result.output == "You must be logged in to perform this operation\n"

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.get_details')
    def test_view_job(self, get_details_mock, is_logged_in_mock):
        """
        Tests a successful job view command
        Returns:

        """
        is_logged_in_mock.return_value = True
        get_details_mock.return_value = MockViewJobResponse
        runner = CliRunner()
        result = runner.invoke(cli, ['job', 'view', 'job_id'])
        get_details_mock.assert_called_once()
        assert result.exit_code == 0
        assert output_message_includes(result, "Status: Done")
        assert output_message_includes(result, "Name: dsfgfdsgs")
        assert output_message_includes(result, "failed: 1")

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.get_details')
    def test_view_job_malformed_id(self, get_details_mock, is_logged_in_mock):
        is_logged_in_mock.return_value = True
        get_details_mock.side_effect = MockBadIdException
        get_details_mock.return_value = MockViewJobResponse
        runner = CliRunner()
        result = runner.invoke(cli, ['job', 'view', 'job_id'])
        get_details_mock.assert_called_once()
        assert result.exit_code == 0
        assert output_message_includes(result, "Bad format for Id")

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.create')
    @patch('pathlib.Path.exists')
    @patch('pathlib.Path.is_dir')
    @patch('pathlib.Path.glob')
    @patch('disco.asset.Asset.upload')
    def test_create_job_input_file(self, asset_upload_mock, path_glob_mock, path_is_dir_mock,
                                    path_exists_mock, job_create_mock, is_logged_in_mock):
        """
        Successful path for creating a job using input files
        Args:
            asset_upload_file_mock:
            path_glob_mock:
            path_is_dir_mock:
            path_exists_mock:
            job_create_mock:

        Returns:

        """
        is_logged_in_mock.return_value = True
        path_exists_mock.return_value = True
        path_is_dir_mock.return_value = False
        job_create_mock.return_value = Job("job_id")
        asset_upload_mock.return_value = 'file_id'
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('script_file.py', 'w') as f:
                f.write("print('hi!')")
                result = runner.invoke(cli, "job create -n job_name -s script_file.py -i input_file".split(' '))
                assert result.exit_code == 0
                assert output_message_includes(result, 'Created job with id job_id')
                job_create_mock.assert_called_with('file_id', ['file_id'], [],
                                                   'job_name',
                                                   's', None, False, upload_requirements_file=True)

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.create')
    @patch('pathlib.Path.exists')
    @patch('pathlib.Path.is_dir')
    @patch('pathlib.Path.glob')
    @patch('disco.asset.Asset.upload')
    def test_create_job_without_req_file(self, asset_upload_mock, path_glob_mock, path_is_dir_mock,
                                   path_exists_mock, job_create_mock, is_logged_in_mock):
        """
        Successful path for creating a job using input files
        Args:
            asset_upload_file_mock:
            path_glob_mock:
            path_is_dir_mock:
            path_exists_mock:
            job_create_mock:

        Returns:

        """
        is_logged_in_mock.return_value = True
        path_exists_mock.return_value = True
        path_is_dir_mock.return_value = False
        job_create_mock.return_value = Job("job_id")
        asset_upload_mock.return_value = 'file_id'
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('script_file.py', 'w') as f:
                f.write("print('hi!')")
                result = runner.invoke(cli, "job create -n job_name -s script_file.py -i "
                                            "input_file --dont-generate-req-file".split(' '))
                assert result.exit_code == 0
                assert output_message_includes(result, 'Created job with id job_id')
                job_create_mock.assert_called_with('file_id', ['file_id'], [],
                                                   'job_name',
                                                   's', None, False, upload_requirements_file=False)


    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.create')
    @patch('pathlib.Path.exists')
    @patch('pathlib.Path.is_dir')
    @patch('pathlib.Path.glob')
    @patch('disco.asset.Asset.upload')
    def test_create_job_constant_file(self, asset_upload_mock, path_glob_mock, path_is_dir_mock,
                                    path_exists_mock, job_create_mock, is_logged_in_mock):
        """
        Successful path for creating a job using constant files
        Args:
            asset_upload_file_mock:
            path_glob_mock:
            path_is_dir_mock:
            path_exists_mock:
            job_create_mock:

        Returns:

        """
        is_logged_in_mock.return_value = True
        path_exists_mock.return_value = True
        path_is_dir_mock.return_value = False
        job_create_mock.return_value = Job("job_id")
        asset_upload_mock.return_value = 'file_id'
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('script_file.py', 'w') as f:
                f.write("print('hi!')")
                result = runner.invoke(cli, "job create -n job_name -s script_file.py -c const_file".split(' '))
                assert result.exit_code == 0
                assert output_message_includes(result, 'Created job with id job_id')
                job_create_mock.assert_called_with('file_id', [], ['file_id'],
                                                   'job_name',
                                                   's', None, False, upload_requirements_file=True)

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.create')
    @patch('pathlib.Path.exists')
    @patch('pathlib.Path.is_dir')
    @patch('pathlib.Path.glob')
    @patch('disco.asset.Asset.upload')
    def test_create_job_many_files(self, asset_upload_mock, path_glob_mock, path_is_dir_mock,
                                      path_exists_mock, job_create_mock, is_logged_in_mock):
        """
        Successful path for creating a job using many input and constant files
        Args:
            asset_upload_file_mock:
            path_glob_mock:
            path_is_dir_mock:
            path_exists_mock:
            job_create_mock:

        Returns:

        """
        is_logged_in_mock.return_value = True
        path_exists_mock.return_value = True
        path_is_dir_mock.return_value = False
        job_create_mock.return_value = Job("job_id")
        asset_upload_mock.return_value = 'file_id'
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('script_file.py', 'w') as f:
                f.write("print('hi!')")
                result = runner.invoke(cli, "job create -n job_name -s script_file.py "
                                            "-i input_file1,input_file2 "
                                            "-c const_file1,const_file2".split(' '))
                assert result.exit_code == 0
                assert output_message_includes(result, 'Created job with id job_id')
                job_create_mock.assert_called_with('file_id', ['file_id', 'file_id'],
                                                   ['file_id', 'file_id'], 'job_name',
                                                   's', None, False, upload_requirements_file=True)

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.create')
    @patch('pathlib.Path.exists')
    @patch('pathlib.Path.is_dir')
    @patch('pathlib.Path.glob')
    @patch('disco.asset.Asset.upload')
    def test_create_job_directory(self, asset_upload_mock, path_glob_mock, path_is_dir_mock,
                                      path_exists_mock, job_create_mock, is_logged_in_mock):
        """
        Successful path for creating a job using directory as input
        Args:
            asset_upload_file_mock:
            path_glob_mock:
            path_is_dir_mock:
            path_exists_mock:
            job_create_mock:

        Returns:

        """
        is_logged_in_mock.return_value = True
        path_exists_mock.return_value = True
        path_is_dir_mock.return_value = True
        job_create_mock.return_value = Job("job_id")
        asset_upload_mock.return_value = 'file_id'
        path_glob_mock.return_value = [Path('file1'), Path('file2')]
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('script_file.py', 'w') as f:
                f.write("print('hi!')")
                result = runner.invoke(cli, "job create -n job_name -s script_file.py "
                                            "-i dir_path ")
                assert result.exit_code == 0
                assert output_message_includes(result, 'Created job with id job_id')
                job_create_mock.assert_called_with('file_id', ['file_id', 'file_id'],
                                                   [], 'job_name',
                                                   's', None, False, upload_requirements_file=True)


    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.create')
    @patch('pathlib.Path.exists')
    @patch('pathlib.Path.is_dir')
    @patch('pathlib.Path.glob')
    @patch('disco.asset.Asset.upload')
    def test_create_job_wildcard(self, asset_upload_mock, path_glob_mock, path_is_dir_mock,
                                      path_exists_mock, job_create_mock, is_logged_in_mock):
        """
        Successful path for creating a job using file with wildcards
        Args:
            asset_upload_file_mock:
            path_glob_mock:
            path_is_dir_mock:
            path_exists_mock:
            job_create_mock:

        Returns:

        """
        script_file_id = self.random_str("script_file_id")
        input_file_id1 = self.random_str("input_file_id1")
        input_file_id2 = self.random_str("input_file_id2")

        is_logged_in_mock.return_value = True
        path_exists_mock.return_value = True
        path_is_dir_mock.return_value = True
        job_create_mock.return_value = Job("job_id")
        asset_upload_mock.side_effect = [script_file_id,  input_file_id1, input_file_id2]
        path_glob_mock.return_value = [Path('file1'), Path('file2')]
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('script_file.py', 'w') as f, open("input_file1.txt", 'w') as i1:
                sleep(1)
                i2 = open("input_file2.txt", 'w')
                input_file_contents1 = self.random_str("input_contents1")
                input_file_contents2 = self.random_str("input_contents2")
                i1.write(input_file_contents1)

                i2.write(input_file_contents2)
                i2.close()
                f.write("print('hi!')")
                result = runner.invoke(cli, "job create -n job_name -s script_file.py "
                                            "-i inp* ")
                assert result.exit_code == 0
                assert output_message_includes(result, 'Created job with id job_id')
                job_create_mock.assert_called_with(script_file_id, [input_file_id1, input_file_id2],
                                                   [], 'job_name',
                                                   's', None, False, upload_requirements_file=True)
                assert asset_upload_mock.call_args_list[1][0] == ('input_file1.txt', Path('input_file1.txt'))
                asset_upload_mock.assert_called_with('input_file2.txt', Path('input_file2.txt'), cluster_id=None)

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.create')
    @patch('pathlib.Path.exists')
    @patch('pathlib.Path.is_dir')
    @patch('pathlib.Path.glob')
    @patch('disco.asset.Asset.upload')
    def test_create_job_not_supported_script(self, asset_upload_mock, path_glob_mock, path_is_dir_mock,
                                      path_exists_mock, job_create_mock, is_logged_in_mock):
        """
        Unsuccessful path for creating a job with a script file that is not supported
        Args:
            asset_upload_file_mock:
            path_glob_mock:
            path_is_dir_mock:
            path_exists_mock:
            job_create_mock:

        Returns:

        """
        is_logged_in_mock.return_value = True
        path_exists_mock.return_value = True
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('script_file', 'w') as f:
                f.write("print('hi!')")
                result = runner.invoke(cli, "job create -n job_name -s script_file")
                assert result.exit_code == 0
                assert output_message_includes(result, 'Cannot use script file. '
                                                       'Currently only Python and '
                                                       'bash scripts are supported')
                job_create_mock.assert_not_called()

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.create')
    @patch('pathlib.Path.exists')
    @patch('pathlib.Path.is_dir')
    @patch('pathlib.Path.glob')
    @patch('disco.asset.Asset.upload')
    def test_create_job_empty_directory(self, asset_upload_mock, path_glob_mock, path_is_dir_mock,
                                  path_exists_mock, job_create_mock, is_logged_in_mock):
        """
        Successful path for creating a job using directory as input
        Args:
            asset_upload_file_mock:
            path_glob_mock:
            path_is_dir_mock:
            path_exists_mock:
            job_create_mock:

        Returns:

        """
        is_logged_in_mock.return_value = True
        path_exists_mock.return_value = True
        path_is_dir_mock.return_value = True
        path_glob_mock.return_value = []
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('script_file.py', 'w') as f:
                f.write("print('hi!')")
                result = runner.invoke(cli, "job create -n job_name -s script_file.py "
                                            "-i dir_path ")
                assert result.exit_code == 0
                assert output_message_includes(result, 'Folder dir_path is empty')
                job_create_mock.assert_not_called()

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.create')
    @patch('pathlib.Path.exists')
    @patch('pathlib.Path.is_dir')
    @patch('pathlib.Path.glob')
    @patch('disco.asset.Asset.upload')
    def test_create_job_missing_input(self, asset_upload_mock, path_glob_mock, path_is_dir_mock,
                                             path_exists_mock, job_create_mock, is_logged_in_mock):
        """
        Unsuccessful path for creating a job with a script file that is not supported
        Args:
            asset_upload_file_mock:
            path_glob_mock:
            path_is_dir_mock:
            path_exists_mock:
            job_create_mock:

        Returns:

        """
        is_logged_in_mock.return_value = True
        path_exists_mock.return_value = False
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('script_file.py', 'w') as f:
                f.write("print('hi!')")
                result = runner.invoke(cli, "job create -n job_name -s script_file.py -i input_file")
                assert result.exit_code == 0
                assert output_message_includes(result, 'input_file doesn\'t exist')
                job_create_mock.assert_not_called()

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.start')
    def test_start_job(self, job_start_mock, is_logged_in_mock):
        is_logged_in_mock.return_value = True
        runner = CliRunner()
        result = runner.invoke(cli, 'job start job_id')
        assert result.exit_code == 0
        assert result.output == 'Job job_id started\n'
        job_start_mock.assert_called()

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.stop')
    def test_stop_job(self, job_stop_mock, is_logged_in_mock):
        is_logged_in_mock.return_value = True
        runner = CliRunner()
        result = runner.invoke(cli, 'job stop job_id')
        assert result.exit_code == 0
        assert result.output == 'Stopping job job_id\n'
        job_stop_mock.assert_called()

    @patch('disco.gql.authentication.Authentication.is_logged_in')
    @patch('disco.Job.archive')
    def test_archive_job(self, job_archive_mock, is_logged_in_mock):
        is_logged_in_mock.return_value = True
        runner = CliRunner()
        result = runner.invoke(cli, 'job archive job_id')
        assert result.exit_code == 0
        assert result.output == 'Job job_id was archived\n'
        job_archive_mock.assert_called()








import click
import cli.utilities.utils as utils
from requests import ConnectionError, HTTPError

from cli.command_utils import error_message
from disco.core.exceptions import NoCredentials


@click.group()
def cli():
    pass

def setup_cli():
    from cli import auth_commands, cluster_commands, job_commands, version_commands
    cli.add_command(auth_commands.login)
    cli.add_command(auth_commands.logout)
    cli.add_command(cluster_commands.cluster)
    cli.add_command(job_commands.job)
    cli.add_command(version_commands.version)

def check_version():
    if utils.is_update_needed():
        click.echo(click.style("There is a newer version. \nPlease upgrade using: pip install disco --upgrade\n", fg='yellow', bold='True'))

def main():
    check_version()
    try:
        setup_cli()
    except Exception as ex:
        print("Error while setting up cli")

    try:
        cli()
    except HTTPError as ex:
        error_message("Unknown error received from server")
    except ConnectionError as ex:
        error_message("Couldn't establish internet connection. "
                      "Please connect to the internet and try again")
    except NoCredentials as ex:
        error_message("You must be logged in to perform this operation")
    except Exception as ex:
        error_message("Unknown error occurred")

if __name__ == '__main__':
    main()





"""
REST queries to the backend
"""
import uuid

import requests

from disco.core.exceptions import GraphQLRequestException
from .request_mixin import _RequestsMixin
from .commands import REGISTERED_COMMANDS
from ..core import utils


class RestQuery(_RequestsMixin):
    """Allows submission of REST queries to the backend and then
    either streaming or getting their output.
    """
    RestMethods = {'get', 'post', 'head', 'put', 'options', 'delete'}

    def __init__(
            self,
            url: str,
            data: dict = None,
            params: dict = None,
            method: str = 'get',
            **kwargs
    ):
        """Builds a request to be sent or streamed.

        Args:
            url: Relative url.
            data: Data to be sent.
            params: Parameters to be sent (see: requests package parameters).
            method: Method  to be used.
            skip_authorisation: If set to `True`,
                won't attempt to authorise the request. (used for login).
            request_id: Optional id to attach to the request and response.
                Defaults to GUID.
        """
        assert method in self.RestMethods
        self.url = url
        self.data = data
        self.params = params
        self.method = method
        # implicit request id on Falsy values
        self.request_id = kwargs.get('request_id') or uuid.uuid4().hex
        self.cookies = kwargs.get('cookies') or {}
        self.headers = kwargs.get('headers') or {}

    @property
    def absolute_path(self):
        """Absolute path for the request.

        Returns:
            str: Url for specific REST url.
        """
        return self.get_rest_endpoint(self.url)

    @_RequestsMixin.handle_500_errors
    def _request(self, stream: bool):
        headers = utils.merge_dictionaries(self.headers, self.disco_headers)
        response = requests.request(
            self.method,
            self.absolute_path,
            params=self.params, cookies=self.cookies or None,
            headers=headers, stream=stream, json=self.data)
        response.raise_for_status()
        response.request_id = self.request_id
        return response

    def stream(self):
        """Stream results rather than getting all at once.

        Returns:
            Raw response bytes.
        """
        ret_data = b''
        with self._request(True) as response:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:  # filter out keep-alive new chunks
                    ret_data += chunk
        return ret_data

    def send(self):
        """Sends the request to the backend and gets the result json.

        Returns:
            The data portion of the response json.
        """
        response = self._request(False)
        json_response = response.json()
        if response.content and 'data' in json_response and \
                json_response['data'] is not None:
            return json_response['data']
        if json_response['errors']:
            raise GraphQLRequestException('Graphql call returned with errors',
                                          json_response['errors'])
        return None

    def __str__(self):
        return f"REST Query: {self.method} {self.url} {self.data}"


class Query(RestQuery):
    """GQL Query.
    The query itself should be stored in `./commands/Q_NAME.gql` and then
    referenced by `Q_NAME` as `operation_name`.
    """
    _rest_properties = {
        'request_id': None
    }

    def __init__(self, operation_name=None, query_string=None, **kwargs):
        """New GQL Query. Only sets up the query but does not execute it.

        Args:
            operation_name: the op_name to be sent and
                the file containing the query itself.
            query_string: optional explicit query string.
                Defaults to the query in the `./commands/operation_name.gql`
            **kwargs: params to be passed as variables to the query.
        """
        base_props = {
            k: kwargs.pop(k, v)
            for k, v in self._rest_properties.items()
        }
        self.operation = operation_name
        self.variables = dict(kwargs)

        self.query_string = query_string
        if self.query_string is None:
            self.query_string = REGISTERED_COMMANDS[operation_name]

        data = utils.filter_nones(
            operationName=self.operation,
            variables=self.variables,
            query=self.query_string
        )
        super().__init__(url='qgl', method='post', data=data, **base_props)

    @property
    def absolute_path(self):
        """Absolute path for the request.

        Returns:
            str: Base url for DISCO GQL requests.
        """
        return self.get_graphql_endpoint()

    def __str__(self):
        return f"GQL Query: " \
            f"{self.operation or self.query_string} " \
            f"{self.variables}"

    def stream(self):
        raise NotImplementedError("Streaming of GQL results is not"
                                  " supported yet")



#  Copyright (c) 2019 Samsung, Inc. and its affiliates.
#
#  This source code is licensed under the RESTRICTED license found in the
#  LICENSE file in the root directory of this source tree.

import io
import os
import pathlib
import uuid

import mock
import pytest

import disco
from .base_test import BaseTest
from disco import Asset
from contextlib import contextmanager

test_file_dir = os.path.dirname(os.path.realpath(__file__))
txt_file = os.path.join(test_file_dir, 'test_files', 'text_file.txt')
img_file = os.path.join(test_file_dir, 'test_files', 'disco-logo.png')

file_content = {}


class TestAsset(BaseTest):

    def test__request_url_for_upload_with_cluster(self):
        asset = Asset()
        fname = self.random_str('filename')
        cluster = self.random_str('cluster')
        excpected = {
            'key': fname,
            'clusterId': cluster
        }
        _rest = mock.MagicMock('Asset.rest')

        asset.rest = _rest
        asset._request_url_for_upload(fname, cluster)
        _rest.assert_called_once_with(url='/files/uploadparams',
                                      data=excpected,
                                      method='post')

    def test__request_url_for_upload_without_cluster(self):
        asset = Asset()
        _rest = mock.MagicMock('Asset.rest')
        asset.rest = _rest

        fname = self.random_str('filename')
        excpected = {
            'key': fname,
        }
        asset._request_url_for_upload(fname, None)
        _rest.assert_called_once_with(url='/files/uploadparams',
                                      data=excpected,
                                      method='post')

    @mock.patch('requests.request')
    def test__upload_file(self, _request: mock.MagicMock):
        asset = Asset()

        url = self.random_str('url')
        file_name = self.random_str('fname')
        form_fields = self.random_dict()
        file_content_bytes = self.random_bytes(100)
        asset._upload(url, file_name, form_fields, file_content_bytes)
        _request.assert_called_once_with(
            'post',
            url,
            data=form_fields,
            files={
                'file': (file_name, self.any_instance_of(io.BytesIO)),
            }
        )
        _request.return_value.raise_for_status.assert_called_once()

    def test__register_file(self):
        asset = Asset()
        _rest = mock.MagicMock('Asset.rest')
        asset.rest = _rest

        token = self.random_str('token')
        send_result = self.random_str('id')
        _rest.return_value = {'id': send_result}
        result = asset._register(token)
        _rest.assert_called_once_with(url='/files',
                                      data={'token': token},
                                      method='post')

        assert result == send_result


def setup_module():
    global file_content
    with io.FileIO(txt_file) as txt_f:
        file_content[FTypes.text] = txt_f.read()

    with io.FileIO(img_file) as img_f:
        file_content[FTypes.image] = img_f.read()


class FTypes(object):
    text = 'Text'
    image = 'Image'


file_argument_parameters = [
    (b'Lol I have text \n and newlines here.', FTypes.text),
    ('Lol I have text \n and newlines here.', FTypes.text),
    (io.StringIO('Lol I have text \n and newlines here.'), FTypes.text),
    (io.BytesIO(b'Lol I have text \n and newlines here.'), FTypes.text),
    (open(os.path.abspath(txt_file), 'r'), FTypes.text),
    (open(os.path.abspath(txt_file), 'rb'), FTypes.text),
    (pathlib.Path(txt_file), FTypes.text),
    (pathlib.Path(img_file), FTypes.image),
    (open(os.path.abspath(img_file), 'rb'), FTypes.image)
]


@pytest.mark.parametrize('file_argument', file_argument_parameters)
def test_upload_file_mocked(file_argument):
    generated_id = uuid.uuid4()
    generated_url = '/some-storage/%s' % generated_id
    generated_name = 'file_%s.bin' % generated_id
    generated_token = 'token>%s' % generated_id
    generated_fields = {'x': generated_id, 'y': generated_name}

    with mock.patch('disco.asset.Asset._request_url_for_upload') \
            as request_url_mock, \
            mock.patch('disco.asset.Asset._upload') as upload_patch_mock, \
            mock.patch('disco.asset.Asset._register') as register_file_mock:
        request_url_mock.return_value = {
            'url': generated_url,
            'fields': generated_fields,
            'token': generated_token
        }

        register_file_mock.return_value = generated_id

        file_id = disco.asset.upload_file(generated_name, file_argument[0])

        request_url_mock.assert_called_once_with(generated_name, None)
        expected_body = file_content[file_argument[1]]

        upload_patch_mock.assert_called_once_with(
            generated_url, generated_name, form_fields=generated_fields,
            file_content_bytes=expected_body)
        register_file_mock.assert_called_once_with(generated_token)

        assert file_id == generated_id



"""
Task Objects
"""
from pathlib import Path
from disco.core.utils import shorten_string


class TaskResult:
    """
    Task Result
    """
    def __init__(self, task_id, raw_result):
        self.task_id = task_id
        self.raw_result = raw_result

    @property
    def stdout(self):
        """The standard output produced by the script."""
        for file_name, content in self.raw_result:
            if file_name.startswith('IqoqoTask.stdout.'):
                return content.strip()

        return None

    def __repr__(self):
        return f"<{type(self).__name__}: " \
            f"job={self.task_id} " \
            f"stdout={shorten_string(self.stdout)}"

    def write_files(self, path):
        """Write all the files created by the job to the local filesystem.

        Args:
            path (str): The path to write the files in.
        """
        path = Path(str(path))
        if path.exists():
            assert path.is_dir()
        for relative_file_path, content in self.raw_result:
            absolute_file_path = path / relative_file_path
            if relative_file_path[-1] == '/':
                absolute_file_path.mkdir(parents=True)
            else:
                absolute_file_path.write_bytes(content)

    def get_raw_result(self):
        """Returns the raw result (file->binary data) of a task"""
        return self.raw_result



"""
Basic Analytics functionality. Exposed `setup_analytics` and `track`
Uses the `python-analytics` package
"""
import json
import os
import uuid

import analytics


def setup_analytics():
    """Sets up analytics tracking"""
    if analytics.write_key:
        return
    if not os.path.exists('.config'):
        return
    with open('.config') as json_file:
        data = json.load(json_file)
        if not data["segment"]:
            return
        analytics.write_key = data["segment"]


def track(event, payload, user_id=str(uuid.uuid4())):
    """Track single event. Refer to `python-analytics` `track` method for more
    info

    Args:
        event:
        payload:
        user_id:
    """
    if analytics.write_key:
        payload["interface"] = "SDK"
        analytics.track(user_id, event, payload)



"""
Upload a file to DISCO, so it could later be used to run jobs.
"""
import io
import pathlib

import requests

from .base_controller import BaseController


class Asset(BaseController):
    """Provides functionality for uploading and downloading disco files"""

    def _request_url_for_upload(self, file_name, cluster_id=None):
        params = {'key': file_name}
        if cluster_id:
            params['clusterId'] = cluster_id
        return self.rest(url='/files/uploadparams',
                         data=params,
                         method='post'
                         )

    @classmethod
    def _upload(cls, url, file_name, form_fields, file_content_bytes):
        response = requests.request(
            'post',
            url,
            data=form_fields,
            files={
                'file': (file_name, io.BytesIO(file_content_bytes)),
            }
        )
        response.raise_for_status()

    def _register(self, token):
        return self.rest(
            url='/files',
            data={'token': token},
            method='post')['id']

    def upload(self, file_name, file, cluster_id=None):
        """Upload a file to DISCO, so it could later be used to run jobs.

        Args:
            file_name (str):
            file: `file` can be either the file contents,
                in binary or string forms, a file
                object, or a Path` object that points to a file.
            cluster_id (str):

        Returns:
            str: The id of the uploaded file.
        """
        if isinstance(file, bytes):
            file_content = file
        elif isinstance(file, str):
            file_content = file.encode()
        elif isinstance(file, pathlib.Path):
            file_content = file.read_bytes()
        elif hasattr(file, 'read'):
            file_content = file.read()
            if isinstance(file_content, bytes):
                pass
            elif isinstance(file_content, str):
                file_content = file_content.encode()
        else:
            file_content = pathlib.Path(str(file)).read_bytes()

        data = self._request_url_for_upload(file_name, cluster_id)
        self._upload(data['url'], file_name, form_fields=data['fields'],
                     file_content_bytes=file_content)
        register_result = self._register(data['token'])
        return register_result


def upload_file(file_name, file, cluster_id=None):
    """Legacy: Uploads file to DISCO, see `Assest.upload` for more info.

    Args:
        file_name (str):
        file: `file` can be either the file contents,
                in binary or string forms, a file
                object, or a Path` object that points to a file.
        cluster_id (str):

    Returns:
        str: The id of the uploaded file.
    """
    return Asset().upload(file_name, file, cluster_id=cluster_id)



"""
Tools for working with zip files in memory
"""
import zipfile as zip_module
import io


def zip_in_memory(files):
    """Zip files in memory and return zip archive as a string.
    Files should be given as tuples of `(file_path, file_contents)`.

    Args:
        files:

    Returns:
        str:
    """
    zip_stream = io.BytesIO()
    with zip_module.ZipFile(
            zip_stream,
            mode='w',
            compression=zip_module.ZIP_DEFLATED
    ) as zip_file:
        assert isinstance(zip_file, zip_module.ZipFile)
        for file_name, file_data in files:
            zip_file.writestr(file_name, file_data)

    return zip_stream.getvalue()


def unzip_in_memory(zip_archive):
    """Unzip a zip archive given as string, returning files
    Files are returned as tuples of `(file_path, file_contents)` .

    Args:
        zip_archive (str):

    Returns:
    """
    zip_stream = io.BytesIO(zip_archive)
    with zip_module.ZipFile(
            zip_stream,
            mode='r',
            compression=zip_module.ZIP_DEFLATED
    ) as zip_file:
        assert isinstance(zip_file, zip_module.ZipFile)
        return tuple((file_name, zip_file.read(file_name)) for file_name in
                     zip_file.namelist())



from tests.base_test import BaseTest
import mock
import disco


class TestClusters(BaseTest):

    @BaseTest.with_config_and_env(authenticated=True)
    def test_list_clusters(self):
        with mock.patch('disco.gql.authentication.Authentication.query') as graphql_mock:
            graphql_mock.return_value = {'fetchProfile': {'defaultClusterId': '000000000000000000000000', 'clusters': [
                {'id': '000000000000000000000000', 'name': 'discoCloud', 'type': 'aws',
                 'cluster': {'region': 'us-west-2', 'externalAccountId': '651133521420',
                             'bucketName': 'disco-store-production'}, 'isLastConnectionCheckValid': True,
                 'isActive': True, 'enableCloudOffload': False},
                {'id': '5d877534fc4015000dd1df18', 'name': 'Itay', 'type': 'aws',
                 'cluster': {'region': 'us-west-1', 'externalAccountId': 'asfasf',
                             'bucketName': 'disco-store-63f7a91c7d7242d8fa3dae62dc249004'},
                 'isLastConnectionCheckValid': False, 'isActive': True, 'enableCloudOffload': False},
                {'id': '5d89d91698e7e4d4c542ad4e', 'name': 'complete-genomics', 'type': 'onPremise',
                 'cluster': {'region': 'us-west-2', 'externalAccountId': 'none',
                             'bucketName': 'disco-store-production'}, 'isLastConnectionCheckValid': True,
                 'isActive': False, 'enableCloudOffload': False}]}}

            result = disco.Cluster.list_clusters()

        assert isinstance(result, list)
        assert len(result) == 3
        assert result[0].id == '000000000000000000000000'
        assert result[0].name == 'discoCloud'
        assert result[0].is_default
        assert result[0].type == 'aws'
        assert result[0].region == 'us-west-2'
        assert result[0].external_account_id == '651133521420'
        assert result[0].bucket_name == 'disco-store-production'
        assert result[0].is_last_connection_check_valid
        assert result[0].is_active
        assert not result[0].cloud_offload_enabled

